@@ -41,8 +41,12 @@ using namespace std;
 #include "global/global_init.h"
 #include "global/signal_handler.h"
 
+#include "perfglue/heap_profiler.h"
+
 #include "include/assert.h"
 
+#include "erasure-code/ErasureCodePlugin.h"
+
 #define dout_subsys ceph_subsys_mon
 
 Monitor *mon = NULL;
@@ -100,24 +104,31 @@ int obtain_monmap(MonitorDBStore &store, bufferlist &bl)
   return -ENOENT;
 }
 
-int mon_data_exists(bool *r)
+int check_mon_data_exists()
 {
   string mon_data = g_conf->mon_data;
   struct stat buf;
   if (::stat(mon_data.c_str(), &buf)) {
-    if (errno == ENOENT) {
-      *r = false;
-    } else {
+    if (errno != ENOENT) {
       cerr << "stat(" << mon_data << ") " << cpp_strerror(errno) << std::endl;
-      return -errno;
     }
-  } else {
-    *r = true;
+    return -errno;
   }
   return 0;
 }
 
-int mon_data_empty(bool *r)
+/** Check whether **mon data** is empty.
+ *
+ * Being empty means mkfs has not been run and there's no monitor setup
+ * at **g_conf->mon_data**.
+ *
+ * If the directory g_conf->mon_data is not empty we will return -ENOTEMPTY.
+ * Otherwise we will return 0.  Any other negative returns will represent
+ * a failure to be handled by the caller.
+ *
+ * @return **0** on success, -ENOTEMPTY if not empty or **-errno** otherwise.
+ */
+int check_mon_data_empty()
 {
   string mon_data = g_conf->mon_data;
 
@@ -128,7 +139,6 @@ int mon_data_empty(bool *r)
   }
   char buf[offsetof(struct dirent, d_name) + PATH_MAX + 1];
 
-  *r = false;
   int code = 0;
   struct dirent *de;
   errno = 0;
@@ -142,7 +152,7 @@ int mon_data_empty(bool *r)
     }
     if (string(".") != de->d_name &&
 	string("..") != de->d_name) {
-      *r = true;
+      code = -ENOTEMPTY;
       break;
     }
   }
@@ -152,14 +162,6 @@ int mon_data_empty(bool *r)
   return code;
 }
 
-int mon_exists(bool *r)
-{
-  int code = mon_data_exists(r);
-  if (code || *r == false)
-    return code;
-  return mon_data_empty(r);
-}
-
 void usage()
 {
   cerr << "usage: ceph-mon -i monid [flags]" << std::endl;
@@ -184,6 +186,21 @@ void usage()
   generic_server_usage();
 }
 
+int preload_erasure_code()
+{
+  string directory = g_conf->osd_pool_default_erasure_code_directory;
+  string plugins = g_conf->osd_erasure_code_plugins;
+  stringstream ss;
+  int r = ErasureCodePluginRegistry::instance().preload(plugins,
+							directory,
+							ss);
+  if (r)
+    derr << ss.str() << dendl;
+  else
+    dout(10) << ss.str() << dendl;
+  return r;
+}
+
 int main(int argc, const char **argv) 
 {
   int err;
@@ -242,6 +259,7 @@ int main(int argc, const char **argv)
 
   global_init(&def_args, args,
               CEPH_ENTITY_TYPE_MON, CODE_ENVIRONMENT_DAEMON, flags);
+  ceph_heap_profiler_init();
 
   uuid_d fsid;
   std::string val;
@@ -290,27 +308,32 @@ int main(int argc, const char **argv)
     usage();
   }
 
-  bool exists;
-  if (mon_exists(&exists))
-    exit(1);
-
-  if (mkfs && exists) {
-    cerr << g_conf->mon_data << " already exists" << std::endl;
-    exit(0);
-  }
-
   // -- mkfs --
   if (mkfs) {
 
-    if (mon_data_exists(&exists))
-      exit(1);
-
-    if (!exists) {
+    int err = check_mon_data_exists();
+    if (err == -ENOENT) {
       if (::mkdir(g_conf->mon_data.c_str(), 0755)) {
 	cerr << "mkdir(" << g_conf->mon_data << ") : "
 	     << cpp_strerror(errno) << std::endl;
 	exit(1);
       }
+    } else if (err < 0) {
+      cerr << "error opening '" << g_conf->mon_data << "': "
+           << cpp_strerror(-err) << std::endl;
+      exit(-err);
+    }
+
+    err = check_mon_data_empty();
+    if (err == -ENOTEMPTY) {
+      // Mon may exist.  Let the user know and exit gracefully.
+      cerr << "'" << g_conf->mon_data << "' already exists and is not empty"
+           << ": monitor may already exist" << std::endl;
+      exit(0);
+    } else if (err < 0) {
+      cerr << "error checking if '" << g_conf->mon_data << "' is empty: "
+           << cpp_strerror(-err) << std::endl;
+      exit(-err);
     }
 
     // resolve public_network -> public_addr
@@ -420,17 +443,60 @@ int main(int argc, const char **argv)
       cerr << argv[0] << ": error creating monfs: " << cpp_strerror(r) << std::endl;
       exit(1);
     }
+    store.close();
     cout << argv[0] << ": created monfs at " << g_conf->mon_data 
 	 << " for " << g_conf->name << std::endl;
     return 0;
   }
 
+  err = check_mon_data_exists();
+  if (err < 0 && err == -ENOENT) {
+    cerr << "monitor data directory at '" << g_conf->mon_data << "'"
+         << " does not exist: have you run 'mkfs'?" << std::endl;
+    exit(1);
+  } else if (err < 0) {
+    cerr << "error accessing monitor data directory at '"
+         << g_conf->mon_data << "': " << cpp_strerror(-err) << std::endl;
+    exit(1);
+  }
+
+  err = check_mon_data_empty();
+  if (err == 0) {
+    derr << "monitor data directory at '" << g_conf->mon_data
+      << "' is empty: have you run 'mkfs'?" << dendl;
+    exit(1);
+  } else if (err < 0 && err != -ENOTEMPTY) {
+    // we don't want an empty data dir by now
+    cerr << "error accessing '" << g_conf->mon_data << "': "
+         << cpp_strerror(-err) << std::endl;
+    exit(1);
+  }
+
+  {
+    // check fs stats. don't start if it's critically close to full.
+    ceph_data_stats_t stats;
+    int err = get_fs_stats(stats, g_conf->mon_data.c_str());
+    if (err < 0) {
+      cerr << "error checking monitor data's fs stats: " << cpp_strerror(err)
+           << std::endl;
+      exit(-err);
+    }
+    if (stats.avail_percent <= g_conf->mon_data_avail_crit) {
+      cerr << "error: monitor data filesystem reached concerning levels of"
+           << " available storage space (available: "
+           << stats.avail_percent << "% " << prettybyte_t(stats.byte_avail)
+           << ")\nyou may adjust 'mon data avail crit' to a lower value"
+           << " to make this go away (default: " << g_conf->mon_data_avail_crit
+           << "%)\n" << std::endl;
+      exit(ENOSPC);
+    }
+  }
+
   // we fork early to prevent leveldb's environment static state from
   // screwing us over
   Preforker prefork;
   if (!(flags & CINIT_FLAG_NO_DAEMON_ACTIONS)) {
-    if (g_conf->daemonize) {
-      global_init_prefork(g_ceph_context, 0);
+    if (global_init_prefork(g_ceph_context, 0) >= 0) {
       prefork.prefork();
       if (prefork.is_parent()) {
 	return prefork.parent_wait();
@@ -439,25 +505,34 @@ int main(int argc, const char **argv)
     }
     common_init_finish(g_ceph_context);
     global_init_chdir(g_ceph_context);
+    if (preload_erasure_code() < 0)
+      prefork.exit(1);
   }
 
   MonitorDBStore *store = new MonitorDBStore(g_conf->mon_data);
 
   Monitor::StoreConverter converter(g_conf->mon_data, store);
   if (store->open(std::cerr) < 0) {
+    int needs_conversion = converter.needs_conversion();
+    if (needs_conversion < 0) {
+      if (needs_conversion == -ENOENT) {
+        derr << "monitor data directory at '" << g_conf->mon_data
+             << "' is not empty but has no valid store nor legacy monitor"
+             << " store." << dendl;
+      } else {
+        derr << "found errors while validating legacy unconverted"
+             << " monitor store: " << cpp_strerror(needs_conversion) << dendl;
+      }
+      prefork.exit(1);
+    }
+
     int ret = store->create_and_open(std::cerr);
     if (ret < 0) {
       derr << "failed to create new leveldb store" << dendl;
       prefork.exit(1);
     }
 
-    ret = converter.needs_conversion();
-    if (ret < 0) {
-      derr << "found errors while validating legacy unconverted monitor store: "
-           << cpp_strerror(ret) << dendl;
-      prefork.exit(1);
-    }
-    if (ret > 0) {
+    if (needs_conversion > 0) {
       dout(0) << "converting monitor store, please do not interrupt..." << dendl;
       int r = converter.convert();
       if (r) {
@@ -521,11 +596,11 @@ int main(int argc, const char **argv)
     ::encode(v, final);
     ::encode(mapbl, final);
 
-    MonitorDBStore::Transaction t;
+    MonitorDBStore::TransactionRef t(new MonitorDBStore::Transaction);
     // save it
-    t.put("monmap", v, mapbl);
-    t.put("monmap", "latest", final);
-    t.put("monmap", "last_committed", v);
+    t->put("monmap", v, mapbl);
+    t->put("monmap", "latest", final);
+    t->put("monmap", "last_committed", v);
     store->apply_transaction(t);
 
     dout(0) << "done." << dendl;
@@ -610,7 +685,7 @@ int main(int argc, const char **argv)
 
   // bind
   int rank = monmap.get_rank(g_conf->name.get_id());
-  Messenger *messenger = Messenger::create(g_ceph_context,
+  Messenger *messenger = Messenger::create(g_ceph_context, g_conf->ms_type,
 					   entity_name_t::MON(rank),
 					   "mon",
 					   0);
@@ -705,6 +780,8 @@ int main(int argc, const char **argv)
 
   messenger->wait();
 
+  store->close();
+
   unregister_async_signal_handler(SIGHUP, sighup_handler);
   unregister_async_signal_handler(SIGINT, handle_mon_signal);
   unregister_async_signal_handler(SIGTERM, handle_mon_signal);
