@@ -465,7 +465,7 @@ void Paxos::handle_last(MMonPaxos *last)
   peer_first_committed[from] = last->first_committed;
   peer_last_committed[from] = last->last_committed;
 
-  if (last->first_committed > last_committed+1) {
+  if (last->first_committed > last_committed + 1) {
     dout(5) << __func__
             << " mon." << from
 	    << " lowest version is too high for our last committed"
@@ -487,7 +487,7 @@ void Paxos::handle_last(MMonPaxos *last)
   for (map<int,version_t>::iterator p = peer_last_committed.begin();
        p != peer_last_committed.end();
        ++p) {
-    if (p->second < first_committed && first_committed > 1) {
+    if (p->second + 1 < first_committed && first_committed > 1) {
       dout(5) << __func__
 	      << " peon " << p->first
 	      << " last_committed (" << p->second
@@ -846,6 +846,12 @@ void Paxos::commit_start()
     state = STATE_WRITING;
   else
     assert(0);
+
+  if (mon->get_quorum().size() > 1) {
+    // cancel timeout event
+    mon->timer.cancel_event(accept_timeout_event);
+    accept_timeout_event = 0;
+  }
 }
 
 void Paxos::commit_finish()
@@ -900,16 +906,12 @@ void Paxos::commit_finish()
 
   if (do_refresh()) {
     commit_proposal();
-
-    finish_contexts(g_ceph_context, waiting_for_commit);
-
     if (mon->get_quorum().size() > 1) {
-      // cancel timeout event
-      mon->timer.cancel_event(accept_timeout_event);
-      accept_timeout_event = 0;
       extend_lease();
     }
 
+    finish_contexts(g_ceph_context, waiting_for_commit);
+
     assert(g_conf->paxos_kill_at != 10);
 
     finish_round();
@@ -1379,7 +1381,7 @@ void Paxos::dispatch(PaxosServiceMessage *m)
 
   case MSG_MON_PAXOS:
     {
-      MMonPaxos *pm = (MMonPaxos*)m;
+      MMonPaxos *pm = reinterpret_cast<MMonPaxos*>(m);
 
       // NOTE: these ops are defined in messages/MMonPaxos.h
       switch (pm->op) {
