@@ -14,7 +14,7 @@ using namespace std;
 #include "global/global_context.h"
 
 #include "Message.h"
-#include "Pipe.h"
+
 #include "messages/MPGStats.h"
 
 #include "messages/MGenericMessage.h"
@@ -66,6 +66,7 @@ using namespace std;
 #include "messages/MOSDSubOp.h"
 #include "messages/MOSDSubOpReply.h"
 #include "messages/MOSDMap.h"
+#include "messages/MMonGetOSDMap.h"
 
 #include "messages/MOSDPGNotify.h"
 #include "messages/MOSDPGQuery.h"
@@ -102,6 +103,7 @@ using namespace std;
 #include "messages/MClientCapRelease.h"
 #include "messages/MClientLease.h"
 #include "messages/MClientSnap.h"
+#include "messages/MClientQuota.h"
 
 #include "messages/MMDSSlaveRequest.h"
 
@@ -135,6 +137,7 @@ using namespace std;
 
 #include "messages/MExportCaps.h"
 #include "messages/MExportCapsAck.h"
+#include "messages/MGatherCaps.h"
 
 
 #include "messages/MDentryUnlink.h"
@@ -376,6 +379,9 @@ Message *decode_message(CephContext *cct, ceph_msg_header& header, ceph_msg_foot
   case CEPH_MSG_MON_GET_MAP:
     m = new MMonGetMap;
     break;
+  case CEPH_MSG_MON_GET_OSDMAP:
+    m = new MMonGetOSDMap;
+    break;
   case CEPH_MSG_MON_GET_VERSION:
     m = new MMonGetVersion();
     break;
@@ -529,6 +535,9 @@ Message *decode_message(CephContext *cct, ceph_msg_header& header, ceph_msg_foot
   case CEPH_MSG_CLIENT_SNAP:
     m = new MClientSnap;
     break;
+  case CEPH_MSG_CLIENT_QUOTA:
+    m = new MClientQuota;
+    break;
 
     // mds
   case MSG_MDS_SLAVE_REQUEST:
@@ -630,6 +639,9 @@ Message *decode_message(CephContext *cct, ceph_msg_header& header, ceph_msg_foot
   case MSG_MDS_EXPORTCAPSACK:
     m = new MExportCapsAck;
     break;
+  case MSG_MDS_GATHERCAPS:
+    m = new MGatherCaps;
+    break;
 
 
   case MSG_MDS_DENTRYUNLINK:
@@ -720,7 +732,7 @@ Message *decode_message(CephContext *cct, ceph_msg_header& header, ceph_msg_foot
       lderr(cct) << "failed to decode message of type " << type
 		 << " v" << header.version
 		 << ": " << e.what() << dendl;
-      ldout(cct, 30) << "dump: \n";
+      ldout(cct, cct->_conf->ms_dump_corrupt_message_level) << "dump: \n";
       m->get_payload().hexdump(*_dout);
       *_dout << dendl;
       if (cct->_conf->ms_die_on_bad_msg)
