@@ -17,6 +17,7 @@
 #ifndef CEPH_MSG_ASYNCCONNECTION_H
 #define CEPH_MSG_ASYNCCONNECTION_H
 
+#include <pthread.h>
 #include <list>
 #include <map>
 using namespace std;
@@ -52,7 +53,7 @@ class AsyncConnection : public Connection {
   int read_until(uint64_t needed, bufferptr &p);
   int _process_connection();
   void _connect();
-  void _stop(bool external=false);
+  void _stop();
   int handle_connect_reply(ceph_msg_connect &connect, ceph_msg_connect_reply &r);
   int handle_connect_msg(ceph_msg_connect &m, bufferlist &aubl, bufferlist &bl);
   void was_session_reset();
@@ -108,8 +109,8 @@ class AsyncConnection : public Connection {
   ostream& _conn_prefix(std::ostream *_dout);
 
   bool is_connected() {
-    // FIXME?
-    return state != STATE_CLOSED;
+    Mutex::Locker l(lock);
+    return state >= STATE_OPEN && state <= STATE_OPEN_TAG_CLOSE;
   }
 
   // Only call when AsyncConnection first construct
@@ -124,10 +125,32 @@ class AsyncConnection : public Connection {
   int send_message(Message *m);
 
   void send_keepalive();
-  // Don't call it from AsyncConnection
+  // mark_down need to ensure all events completely finished. So we introduce
+  // a separator impl:
+  //
+  // Thread A             Event loop N          Event loop N+1
+  // mark_down()
+  // dispatch
+  // wait
+  //  \                   C_handle_stop
+  //  \                   stop()
+  //  \                   _stop()
+  //  \                   dispatch
+  //  \                                         C_handle_signal
+  //  \                                         wakeup_stop()
+  //  \                                         signal
+  // finished
+  //
+  // The above flow only happen when the caller isn't the pthread own center,
+  // if the owner of center is self, it's safe to call _stop() directly;
   void mark_down() {
-    Mutex::Locker l(lock);
-    _stop(true);
+    Mutex::Locker l(stop_lock);
+    if (center->get_owner() == pthread_self()) {
+      stop();
+    } else {
+      center->dispatch_event_external(stop_handler);
+      stop_cond.Wait(stop_lock);
+    }
   }
   void mark_disposable() {
     Mutex::Locker l(lock);
@@ -222,6 +245,7 @@ class AsyncConnection : public Connection {
   Messenger::Policy policy;
   map<int, list<Message*> > out_q;  // priority queue for outbound msgs
   list<Message*> sent;
+  list<Message*> local_messages;    // local deliver
   Mutex lock;
   utime_t backoff;         // backoff time
   bool open_write;
@@ -229,8 +253,18 @@ class AsyncConnection : public Connection {
   EventCallbackRef write_handler;
   EventCallbackRef reset_handler;
   EventCallbackRef remote_reset_handler;
+  EventCallbackRef connect_handler;
+  EventCallbackRef fast_connect_handler;
+  EventCallbackRef accept_handler;
+  EventCallbackRef fast_accept_handler;
+  EventCallbackRef stop_handler;
+  EventCallbackRef signal_handler;
+  EventCallbackRef local_deliver_handler;
   bool keepalive;
   struct iovec msgvec[IOV_LEN];
+  Mutex stop_lock; // used to protect `mark_down_cond`
+  Cond stop_cond;
+  set<uint64_t> register_time_events; // need to delete it if stop
 
   // Tis section are temp variables used by state transition
 
@@ -250,6 +284,12 @@ class AsyncConnection : public Connection {
   // Accepting state
   entity_addr_t socket_addr;
   CryptoKey session_key;
+  bool replacing;    // when replacing process happened, we will reply connect
+                     // side with RETRY tag and accept side will clear replaced
+                     // connection. So when connect side reissue connect_msg,
+                     // there won't exists conflicting connection so we use
+                     // "replacing" to skip RESETSESSION to avoid detect wrong
+                     // presentation
 
   // used only for local state, it will be overwrite when state transition
   bufferptr state_buffer;
@@ -264,6 +304,17 @@ class AsyncConnection : public Connection {
   // used by eventcallback
   void handle_write();
   void process();
+  void wakeup_from(uint64_t id);
+  // Helper: only called by C_handle_stop
+  void stop() {
+    Mutex::Locker l(lock);
+    _stop();
+  }
+  void wakeup_stop() {
+    Mutex::Locker l(stop_lock);
+    stop_cond.Signal();
+  }
+  void local_deliver();
 }; /* AsyncConnection */
 
 typedef boost::intrusive_ptr<AsyncConnection> AsyncConnectionRef;
