@@ -13,7 +13,7 @@
  */
 
 #include <boost/variant.hpp>
-#include <boost/optional/optional_io.hpp>
+#include <boost/optional.hpp>
 #include <iostream>
 #include <sstream>
 
@@ -21,6 +21,9 @@
 #include "ECBackend.h"
 #include "messages/MOSDPGPush.h"
 #include "messages/MOSDPGPushReply.h"
+#include "ReplicatedPG.h"
+
+class ReplicatedPG;
 
 #define dout_subsys ceph_subsys_osd
 #define DOUT_PREFIX_ARGS this
@@ -213,8 +216,8 @@ struct RecoveryMessages {
     const hobject_t &hoid, uint64_t off, uint64_t len,
     const set<pg_shard_t> &need,
     bool attrs) {
-    list<pair<uint64_t, uint64_t> > to_read;
-    to_read.push_back(make_pair(off, len));
+    list<boost::tuple<uint64_t, uint64_t, uint32_t> > to_read;
+    to_read.push_back(boost::make_tuple(off, len, 0));
     assert(!reads.count(hoid));
     reads.insert(
       make_pair(
@@ -832,6 +835,11 @@ void ECBackend::handle_sub_write(
     op.trim_rollback_to,
     !(op.t.empty()),
     localt);
+
+  if (!(dynamic_cast<ReplicatedPG *>(get_parent())->is_undersized()) &&
+      get_parent()->whoami_shard().shard >= ec_impl->get_data_chunk_count())
+    op.t.set_fadvise_flag(CEPH_OSD_OP_FLAG_FADVISE_DONTNEED);
+
   localt->append(op.t);
   if (on_local_applied_sync) {
     dout(10) << "Queueing onreadable_sync: " << on_local_applied_sync << dendl;
@@ -856,11 +864,11 @@ void ECBackend::handle_sub_read(
   ECSubRead &op,
   ECSubReadReply *reply)
 {
-  for(map<hobject_t, list<pair<uint64_t, uint64_t> > >::iterator i =
+  for(map<hobject_t, list<boost::tuple<uint64_t, uint64_t, uint32_t> > >::iterator i =
         op.to_read.begin();
       i != op.to_read.end();
       ++i) {
-    for (list<pair<uint64_t, uint64_t> >::iterator j = i->second.begin();
+    for (list<boost::tuple<uint64_t, uint64_t, uint32_t> >::iterator j = i->second.begin();
 	 j != i->second.end();
 	 ++j) {
       bufferlist bl;
@@ -868,9 +876,9 @@ void ECBackend::handle_sub_read(
 	i->first.is_temp() ? temp_coll : coll,
 	ghobject_t(
 	  i->first, ghobject_t::NO_GEN, get_parent()->whoami_shard().shard),
-	j->first,
-	j->second,
-	bl,
+	j->get<0>(),
+	j->get<1>(),
+	bl, j->get<2>(),
 	false);
       if (r < 0) {
 	assert(0);
@@ -880,7 +888,7 @@ void ECBackend::handle_sub_read(
       } else {
 	reply->buffers_read[i->first].push_back(
 	  make_pair(
-	    j->first,
+	    j->get<0>(),
 	    bl)
 	  );
       }
@@ -949,7 +957,7 @@ void ECBackend::handle_sub_read_reply(
       // We canceled this read! @see filter_read_op
       continue;
     }
-    list<pair<uint64_t, uint64_t> >::const_iterator req_iter =
+    list<boost::tuple<uint64_t, uint64_t, uint32_t> >::const_iterator req_iter =
       rop.to_read.find(i->first)->second.to_read.begin();
     list<
       boost::tuple<
@@ -962,7 +970,7 @@ void ECBackend::handle_sub_read_reply(
       assert(riter != rop.complete[i->first].returned.end());
       pair<uint64_t, uint64_t> adjusted =
 	sinfo.aligned_offset_len_to_chunk(
-	  *req_iter);
+	  make_pair(req_iter->get<0>(), req_iter->get<1>()));
       assert(adjusted.first == j->first);
       riter->get<2>()[from].claim(j->second);
     }
@@ -1396,22 +1404,23 @@ void ECBackend::start_read_op(
       op.obj_to_source[i->first].insert(*j);
       op.source_to_obj[*j].insert(i->first);
     }
-    for (list<pair<uint64_t, uint64_t> >::const_iterator j =
+    for (list<boost::tuple<uint64_t, uint64_t, uint32_t> >::const_iterator j =
 	   i->second.to_read.begin();
 	 j != i->second.to_read.end();
 	 ++j) {
       reslist.push_back(
 	boost::make_tuple(
-	  j->first,
-	  j->second,
+	  j->get<0>(),
+	  j->get<1>(),
 	  map<pg_shard_t, bufferlist>()));
       pair<uint64_t, uint64_t> chunk_off_len =
-	sinfo.aligned_offset_len_to_chunk(
-	  *j);
+	sinfo.aligned_offset_len_to_chunk(make_pair(j->get<0>(), j->get<1>()));
       for (set<pg_shard_t>::const_iterator k = i->second.need.begin();
 	   k != i->second.need.end();
 	   ++k) {
-	messages[*k].to_read[i->first].push_back(chunk_off_len);
+	messages[*k].to_read[i->first].push_back(boost::make_tuple(chunk_off_len.first,
+								    chunk_off_len.second,
+								    j->get<2>()));
       }
       assert(!need_attrs);
     }
@@ -1570,6 +1579,7 @@ int ECBackend::objects_read_sync(
   const hobject_t &hoid,
   uint64_t off,
   uint64_t len,
+  uint32_t op_flags,
   bufferlist *bl)
 {
   return -EOPNOTSUPP;
@@ -1579,12 +1589,12 @@ struct CallClientContexts :
   public GenContext<pair<RecoveryMessages*, ECBackend::read_result_t& > &> {
   ECBackend *ec;
   ECBackend::ClientAsyncReadStatus *status;
-  list<pair<pair<uint64_t, uint64_t>,
+  list<pair<boost::tuple<uint64_t, uint64_t, uint32_t>,
 	    pair<bufferlist*, Context*> > > to_read;
   CallClientContexts(
     ECBackend *ec,
     ECBackend::ClientAsyncReadStatus *status,
-    const list<pair<pair<uint64_t, uint64_t>,
+    const list<pair<boost::tuple<uint64_t, uint64_t, uint32_t>,
 		    pair<bufferlist*, Context*> > > &to_read)
     : ec(ec), status(status), to_read(to_read) {}
   void finish(pair<RecoveryMessages *, ECBackend::read_result_t &> &in) {
@@ -1592,12 +1602,12 @@ struct CallClientContexts :
     assert(res.returned.size() == to_read.size());
     assert(res.r == 0);
     assert(res.errors.empty());
-    for (list<pair<pair<uint64_t, uint64_t>,
+    for (list<pair<boost::tuple<uint64_t, uint64_t, uint32_t>,
 		   pair<bufferlist*, Context*> > >::iterator i = to_read.begin();
 	 i != to_read.end();
 	 to_read.erase(i++)) {
       pair<uint64_t, uint64_t> adjusted =
-	ec->sinfo.offset_len_to_stripe_bounds(i->first);
+	ec->sinfo.offset_len_to_stripe_bounds(make_pair(i->first.get<0>(), i->first.get<1>()));
       assert(res.returned.front().get<0>() == adjusted.first &&
 	     res.returned.front().get<1>() == adjusted.second);
       map<int, bufferlist> to_decode;
@@ -1617,8 +1627,8 @@ struct CallClientContexts :
       assert(i->second.first);
       i->second.first->substr_of(
 	bl,
-	i->first.first - adjusted.first,
-	MIN(i->first.second, bl.length() - (i->first.first - adjusted.first)));
+	i->first.get<0>() - adjusted.first,
+	MIN(i->first.get<1>(), bl.length() - (i->first.get<0>() - adjusted.first)));
       if (i->second.second) {
 	i->second.second->complete(i->second.first->length());
       }
@@ -1636,7 +1646,7 @@ struct CallClientContexts :
     }
   }
   ~CallClientContexts() {
-    for (list<pair<pair<uint64_t, uint64_t>,
+    for (list<pair<boost::tuple<uint64_t, uint64_t, uint32_t>,
 		   pair<bufferlist*, Context*> > >::iterator i = to_read.begin();
 	 i != to_read.end();
 	 to_read.erase(i++)) {
@@ -1647,21 +1657,23 @@ struct CallClientContexts :
 
 void ECBackend::objects_read_async(
   const hobject_t &hoid,
-  const list<pair<pair<uint64_t, uint64_t>,
+  const list<pair<boost::tuple<uint64_t, uint64_t, uint32_t>,
 		  pair<bufferlist*, Context*> > > &to_read,
   Context *on_complete)
 {
   in_progress_client_reads.push_back(ClientAsyncReadStatus(on_complete));
   CallClientContexts *c = new CallClientContexts(
     this, &(in_progress_client_reads.back()), to_read);
-  list<pair<uint64_t, uint64_t> > offsets;
-  for (list<pair<pair<uint64_t, uint64_t>,
+
+  list<boost::tuple<uint64_t, uint64_t, uint32_t> > offsets;
+  pair<uint64_t, uint64_t> tmp;
+  for (list<pair<boost::tuple<uint64_t, uint64_t, uint32_t>,
 		 pair<bufferlist*, Context*> > >::const_iterator i =
 	 to_read.begin();
        i != to_read.end();
        ++i) {
-    offsets.push_back(
-      sinfo.offset_len_to_stripe_bounds(i->first));
+    tmp = sinfo.offset_len_to_stripe_bounds(make_pair(i->first.get<0>(), i->first.get<1>()));
+    offsets.push_back(boost::make_tuple(tmp.first, tmp.second, i->first.get<2>()));
   }
 
   const vector<int> &chunk_mapping = ec_impl->get_chunk_mapping();
