@@ -1,3 +1,6 @@
+// -*- mode:C++; tab-width:8; c-basic-offset:2; indent-tabs-mode:t -*-
+// vim: ts=8 sw=2 smarttab
+
 #include "rgw_cache.h"
 
 #include <errno.h>
@@ -6,7 +9,7 @@
 
 using namespace std;
 
-int ObjectCache::get(string& name, ObjectCacheInfo& info, uint32_t mask)
+int ObjectCache::get(string& name, ObjectCacheInfo& info, uint32_t mask, rgw_cache_entry_info *cache_info)
 {
   RWLock::RLocker l(lock);
 
@@ -17,16 +20,25 @@ int ObjectCache::get(string& name, ObjectCacheInfo& info, uint32_t mask)
     return -ENOENT;
   }
 
-  ObjectCacheEntry& entry = iter->second;
+  ObjectCacheEntry *entry = &iter->second;
 
-  if (lru_counter - entry.lru_promotion_ts > lru_window) {
-    ldout(cct, 20) << "cache get: touching lru, lru_counter=" << lru_counter << " promotion_ts=" << entry.lru_promotion_ts << dendl;
+  if (lru_counter - entry->lru_promotion_ts > lru_window) {
+    ldout(cct, 20) << "cache get: touching lru, lru_counter=" << lru_counter << " promotion_ts=" << entry->lru_promotion_ts << dendl;
     lock.unlock();
     lock.get_write(); /* promote lock to writer */
 
+    /* need to redo this because entry might have dropped off the cache */
+    iter = cache_map.find(name);
+    if (iter == cache_map.end()) {
+      ldout(cct, 10) << "lost race! cache get: name=" << name << " : miss" << dendl;
+      if(perfcounter) perfcounter->inc(l_rgw_cache_miss);
+      return -ENOENT;
+    }
+
+    entry = &iter->second;
     /* check again, we might have lost a race here */
-    if (lru_counter - entry.lru_promotion_ts > lru_window) {
-      touch_lru(name, entry, iter->second.lru_iter);
+    if (lru_counter - entry->lru_promotion_ts > lru_window) {
+      touch_lru(name, *entry, iter->second.lru_iter);
     }
   }
 
@@ -39,12 +51,59 @@ int ObjectCache::get(string& name, ObjectCacheInfo& info, uint32_t mask)
   ldout(cct, 10) << "cache get: name=" << name << " : hit" << dendl;
 
   info = src;
+  if (cache_info) {
+    cache_info->cache_locator = name;
+    cache_info->gen = entry->gen;
+  }
   if(perfcounter) perfcounter->inc(l_rgw_cache_hit);
 
   return 0;
 }
 
-void ObjectCache::put(string& name, ObjectCacheInfo& info)
+bool ObjectCache::chain_cache_entry(list<rgw_cache_entry_info *>& cache_info_entries, RGWChainedCache::Entry *chained_entry)
+{
+  RWLock::WLocker l(lock);
+
+  list<rgw_cache_entry_info *>::iterator citer;
+
+  list<ObjectCacheEntry *> cache_entry_list;
+
+  /* first verify that all entries are still valid */
+  for (citer = cache_info_entries.begin(); citer != cache_info_entries.end(); ++citer) {
+    rgw_cache_entry_info *cache_info = *citer;
+
+    ldout(cct, 10) << "chain_cache_entry: cache_locator=" << cache_info->cache_locator << dendl;
+    map<string, ObjectCacheEntry>::iterator iter = cache_map.find(cache_info->cache_locator);
+    if (iter == cache_map.end()) {
+      ldout(cct, 20) << "chain_cache_entry: couldn't find cachce locator" << dendl;
+      return false;
+    }
+
+    ObjectCacheEntry *entry = &iter->second;
+
+    if (entry->gen != cache_info->gen) {
+      ldout(cct, 20) << "chain_cache_entry: entry.gen (" << entry->gen << ") != cache_info.gen (" << cache_info->gen << ")" << dendl;
+      return false;
+    }
+
+    cache_entry_list.push_back(entry);
+  }
+
+
+  chained_entry->cache->chain_cb(chained_entry->key, chained_entry->data);
+
+  list<ObjectCacheEntry *>::iterator liter;
+
+  for (liter = cache_entry_list.begin(); liter != cache_entry_list.end(); ++liter) {
+    ObjectCacheEntry *entry = *liter;
+
+    entry->chained_entries.push_back(make_pair<RGWChainedCache *, string>(chained_entry->cache, chained_entry->key));
+  }
+
+  return true;
+}
+
+void ObjectCache::put(string& name, ObjectCacheInfo& info, rgw_cache_entry_info *cache_info)
 {
   RWLock::WLocker l(lock);
 
@@ -59,6 +118,15 @@ void ObjectCache::put(string& name, ObjectCacheInfo& info)
   ObjectCacheEntry& entry = iter->second;
   ObjectCacheInfo& target = entry.info;
 
+  for (list<pair<RGWChainedCache *, string> >::iterator iiter = entry.chained_entries.begin();
+       iiter != entry.chained_entries.end(); ++iiter) {
+    RGWChainedCache *chained_cache = iiter->first;
+    chained_cache->invalidate(iiter->second);
+  }
+
+  entry.chained_entries.clear();
+  entry.gen++;
+
   touch_lru(name, entry, entry.lru_iter);
 
   target.status = info.status;
@@ -70,6 +138,11 @@ void ObjectCache::put(string& name, ObjectCacheInfo& info)
     return;
   }
 
+  if (cache_info) {
+    cache_info->cache_locator = name;
+    cache_info->gen = entry.gen;
+  }
+
   target.flags |= info.flags;
 
   if (info.flags & CACHE_FLAG_META)
@@ -111,6 +184,13 @@ void ObjectCache::remove(string& name)
     return;
 
   ldout(cct, 10) << "removing " << name << " from cache" << dendl;
+  ObjectCacheEntry& entry = iter->second;
+
+  for (list<pair<RGWChainedCache *, string> >::iterator iiter = entry.chained_entries.begin();
+       iiter != entry.chained_entries.end(); ++iiter) {
+    RGWChainedCache *chained_cache = iiter->first;
+    chained_cache->invalidate(iiter->second);
+  }
 
   remove_lru(name, iter->second.lru_iter);
   cache_map.erase(iter);
